!Brendel-Bormann model for the computation of the relative permittivity of silver.
!Other materials can be added
!
!Original code (namely eps_BB.m) written in Matlab  
!Relative Permittivity of various metals originally coded by Collin
!Meierbachtol (C)2009 based on the Brendel-Bormann model described by 
!A. D. Rakic, et. al., App. Opt., vol. 37, no. 22, 1998.
!The original file may be downloaded from
!https://www.mathworks.com/matlabcentral/fileexchange/26025-permittivity-of-metals-at-optical-wavelengths--brendel-bormann-
!
!Fortran adaptation and HOAD by Isabelle Charpentier

!Equation number from "Optimized higher-order automatic differentiation for the Faddeeva function"

SUBROUTINE BBmodel(lambda,material,epsilon,ntilde)
USE HOAD_types; USE HOAD_add;  USE HOAD_div;    USE HOAD_prod; 
USE HOAD_power; USE HOAD_sub;  USE HOAD_uminus; USE HOAD_sqrt;
USE HOAD_real;  USE HOAD_aimag; USE HOAD_faddeeva

IMPLICIT NONE
TYPE(Real8d), INTENT(IN)    :: lambda         !wavelength
TYPE(Complex16d),INTENT(OUT):: epsilon,ntilde !dielectric function, complex refractive index
 CHARACTER*3 , INTENT(IN)   :: material       !silver (other materials may be added)

!Physical constants
DOUBLE PRECISION::pi
PARAMETER(pi=3.14159265358979323846d0) 
DOUBLE PRECISION::c    !Speed of light [m/s]
PARAMETER (c=2.997d8)
DOUBLE PRECISION::h    !Planck's constant [J s]
PARAMETER (h=6.626d-34)
DOUBLE PRECISION::hbar !Normalizad Planck's constant [J s]
PARAMETER (hbar=h/(2*pi))
DOUBLE PRECISION::q    !Quantized electronic charge
PARAMETER (q=1.602d-19)

!Brendel-Bormann matrix elements for fj, GAMMAj, wj, and sigmaj
DOUBLE PRECISION::f0,Gamma0,omegap
DOUBLE PRECISION,DIMENSION(5) :: f,Gamma,omega,sigma
!local variables
TYPE(Real8d)::omegalight            !omegalight=2*pi*c/lambda
TYPE(Real8d)::epsilon_Re,epsilon_Im !real and imaginary parts of the dielectric function
INTEGER::j                          !loop index over the Brendel-Bormann oscillators
TYPE(Complex16d)::zplus,zminus      !arguments of the Faddeeva function in equation (20)
TYPE(Complex16d)::cefzplus,cefzminus!results of the Faddeeva function at point zplus and zminus
TYPE(Complex16d)::epsf              !Intraband contribution  Equation (18)   
TYPE(Complex16d)::alphaj            !Used in equation (20)       

SELECT CASE (material)
  CASE ('Ag')
        !Plasma frequency
        omegap = 9.01d0*(q/hbar);
        !Oscillators' strength
        f0   = 0.821d0
        f(:) =     [0.05d0, 0.133d0, 0.051d0, 0.467d0, 4d0];
        !Damping frequency of each oscillator
        Gamma0   = 0.049d0*(q/hbar)
        Gamma(:) = [0.074d0, 0.035d0, 0.083d0, 0.125d0, 0.179d0]*(q/hbar); 
        !Resonant frequency of each oscillator
        omega(:) = [2.025d0, 5.185d0, 4.343d0, 9.809d0, 18.56d0]*(q/hbar); 
        !
        sigma(:) = [1.894d0, 0.665d0, 0.189d0, 1.17d0, 0.516d0]*(q/hbar);
END SELECT

omegalight=2*pi*c/lambda
   
!Intraband contribution  Equation (18)   
   epsf=1d0 - f0*omegap**2/(omegalight*(omegalight+(0.d0,1.d0)*Gamma0))

!Second (summation) component of relative permittivity
!Total permittivity calculation
epsilon=epsf
DO j=1,5                            !Loops related to equation (19)
!Compute arguments for the calls to Faddeeva function equation (20)
   alphaj=sqrt(omegalight**2+(0.d0,1.d0)*omegalight*Gamma(j))
   zplus =(alphaj+omega(j))/(sqrt(2d0)*sigma(j))
   zminus=(alphaj-omega(j))/(sqrt(2d0)*sigma(j))
!Call to the faddeeva function
   CALL faddeeva(zplus,cefzplus)
   CALL faddeeva(zminus,cefzminus)
   epsilon = epsilon + ((0.d0,1.d0)*sqrt(pi)*f(j)*omegap**2/(2*sqrt(2.d0)*alphaj*sigma(j)))*(cefzplus+cefzminus)
END DO

! Real part of dielectric constant
epsilon_Re= REAL(epsilon);
! Imaginary part of dielectric constant
epsilon_Im=AIMAG(epsilon)
! Complex refractive index [2]: N = n + i*k    !Equation (21)
ntilde = (sqrt(sqrt((epsilon_Re)**2 + (epsilon_Im)**2) + epsilon_Re) &
         + (0.d0,1.d0)*sqrt(sqrt((epsilon_Re)**2 + (epsilon_Im)**2) - epsilon_Re))/sqrt(2.d0);
END SUBROUTINE BBmodel

!Copyright (c) 2016, CNRS and University of Strasbourg, France
!All rights reserved.

!Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

!1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

!2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

!3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

!THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
