!Higher order automatic differentiation up to order Ck for the 2F1 function implemented in hygfz.for
!where Ck is a global variable defining the current truncation order of differentiation
!
!The HOAD_2F1 module allows to package the different versions of 2F1 so that they can be imported into  program units to be automatically differentiated at higher order.
!
!2F1 is overloaded for complex numbers and Complex16d Taylor series. 
!
!Author: I. Charpentier

!External mathematical function
!    hygfz.for
!from https://people.sc.fsu.edu/~jburkardt/f77_src/special_functions/mhygfz.for

MODULE HOAD_2F1
  USE HOAD_types
  INTERFACE fun_2F1
!Overloaded 2F1 
    MODULE PROCEDURE fun_2F1_complex16  !Call of hygfz with a complex number argument
    MODULE PROCEDURE fun_2F1_complex16d !HOAD based on the second order ODE formula
  END INTERFACE fun_2F1
CONTAINS 

!The choice has been to let the external mathematical function implementation outside the module

!-----
!Interface between:
!  * the effective HOAD for the 2F1 function and 
!  * the implementation of the 2F1 function, here the hygfz.for
!
  SUBROUTINE fun_2F1_complex16 (a,b,c,z,w)
    IMPLICIT NONE 
    REAL*8, INTENT(IN):: a,b,c  !parameters
    COMPLEX*16, INTENT(IN) ::z
    COMPLEX*16, INTENT(OUT)::w
    CALL hygfz(a,b,c,z,w)
  END SUBROUTINE fun_2F1_complex16

!-----
!OVERLOADING w=2F1(a,b,c;z): THE ARGUMENT IS A Complex16d TAYLOR SERIES  
  SUBROUTINE fun_2F1_Complex16d (a2F1,b2F1,c2F1,z,w)
    USE HOAD_types; USE HOAD_add;  USE HOAD_div;    USE HOAD_prod; 
    USE HOAD_sub;  USE HOAD_uminus; 
    IMPLICIT NONE 
    REAL*8::a2F1,b2F1,c2F1              !parameters
    TYPE(Complex16d), INTENT(IN)::z
    TYPE(Complex16d), INTENT(OUT)::w
    !Local variables
    DOUBLE PRECISION::pi
    PARAMETER (pi=3.14159265358979323846d0) 
    TYPE(Complex16d)::alpha,beta
    REAL*8::gamma
    TYPE(Complex16d)::z0,z1,z2,w0,w1,z12,z13,d,g,a
    COMPLEX*16::w2 
    INTEGER :: j

    !Initialization
    z0=z; z1=z0; z2=z0;  
    DO Ck=0,Ckmax-1; 
      z1%d(Ck)=z0%d(Ck+1)*(Ck+1); ENDDO !scaling
    DO Ck=0,Ckmax-2; 
      z2%d(Ck)=z0%d(Ck+2)*(Ck+1)*(Ck+2); ENDDO
    w=z*0.d0; w0=w; w1=w; 

    !Mathematical function phi
    CALL fun_2F1(a2F1,b2F1,c2F1,z%d(0),w%d(0)); 
    CALL fun_2F1(a2F1+1,b2F1+1,c2F1+1,z%d(0),w%d(1));
    w%d(1)=(a2F1*b2F1/c2F1)*w%d(1)*z%d(1);
    alpha=z*(1-z); beta=c2F1-(a2F1+b2F1+1)*z; 
    gamma=-a2F1*b2F1 
 
    !Intermediate variables 
    z12=z1*z1; z13=z12*z1;
    d=1.d0/(alpha*z1);
    g=-gamma*z13*d; 
    a=(alpha*z2-beta*z12)*d; 

    !Generic AD of phi
    w0%d(0) = w%d(0);
    w0%d(1) = w%d(1); w1%d(0) =w%d(1);
    DO Ck=2,Ckmax
      w2=(0.d0,0.d0)
      DO j=0,Ck-2
        w2=w2+g%d(j)*w0%d(Ck-2-j)+a%d(j)*w1%d(Ck-2-j)
      ENDDO
      w0%d(Ck)=w2/((Ck-1)*Ck);   !unscaling
      w1%d(Ck-1)=w0%d(Ck)*Ck;
      w%d(Ck)=w0%d(Ck);
    ENDDO
    Ck=Ckmax 
  END SUBROUTINE fun_2F1_complex16d
END MODULE HOAD_2F1

!Copyright (c) 2016, CNRS and University of Strasbourg, France
!All rights reserved.

!Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

!1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

!2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

!3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

!THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
