!Higher order automatic differentiation up to order Ck for the ChebyshevTn function implemented with othpl.for
!where Ck is a global variable defining the current truncation order of differentiation
!
!The HOAD_ChebyshevTn module may be imported into program units to be automatically differentiated at higher order.
!
!ChebyshevTn is overloaded for real8d variables. 
!
!External mathematical function
!    othpl.for    from
!https://people.sc.fsu.edu/~jburkardt/f77_src/special_functions/special_functions.f
!
!the othpl routine is copyrighted by Shanjie Zhang and Jianming Jin.  However, 
!they give permission to incorporate this routine into a user program 
!provided that the copyright is acknowledged.
!Reference:
!Shanjie Zhang, Jianming Jin, Computation of Special Functions, Wiley, 1996,ISBN: 0-471-11963-6, LC: QA351.C45.

!Author: I. Charpentier

MODULE HOAD_ChebyshevTn
  USE HOAD_types
  INTERFACE ChebyshevTn
!Overloaded ChebyshevTn 
    MODULE PROCEDURE ChebyshevTn_Real8  !Call of othpl with KF=1
    MODULE PROCEDURE ChebyshevTn_Real8d !HOAD based on the second order ODE formula
  END INTERFACE ChebyshevTn
CONTAINS 

!The choice has been to let the external mathematical function implementation outside the module

!-----
!Interface between:
!  * the effective HOAD for the ChebyshevTn function and 
!  * the implementation of the ChebyshevTn function, here the routine othpl.for
!
  SUBROUTINE ChebyshevTn_Real8 (KF,n,x,PL,DPL)
    IMPLICIT NONE 
    INTEGER, INTENT(IN)::KF
    INTEGER, INTENT(IN)::n     !Order of orthogonal polynomials
    REAL*8, INTENT(IN) ::x      !Argument of orthogonal polynomials
    REAL*8, INTENT(OUT)::PL(0:n)  !Tn(x)
    REAL*8, INTENT(OUT)::DPL(0:n) !Tn'(x)
    !KF=1 for Chebyshev polynomial Tn(x)   
    !KF=2 for Chebyshev polynomial Un(x)
    CALL othpl(KF,n,x,PL,DPL)
  END SUBROUTINE ChebyshevTn_Real8
!-----
!OVERLOADING v=Tn(x): THE ARGUMENT x IS A Real8d TAYLOR SERIES, n IS AN INTEGER
  SUBROUTINE ChebyshevTn_Real8d (n,x,v) 
    USE HOAD_types; USE HOAD_add;  USE HOAD_div;    USE HOAD_prod; 
    USE HOAD_sub;  USE HOAD_uminus; 
    IMPLICIT NONE 
    INTEGER, INTENT(IN)::n
    TYPE(Real8d), INTENT(IN):: x
    TYPE(Real8d):: v
    !Local variables
    TYPE(Real8d)::alpha,beta 
    REAL*8::gamma
    TYPE(Real8d)::x0,x1,x2,v0,v1,x12,x13,d,g,a
    REAL*8::v2
    REAL*8::PL(0:n),DPL(0:n)
    INTEGER :: j,KF

    !Initialization
    x0=x; x1=x0; x2=x0;  
    DO Ck=0,Ckmax-1; 
      x1%d(Ck)=x0%d(Ck+1)*(Ck+1); ENDDO !scaling
    DO Ck=0,Ckmax-2; 
      x2%d(Ck)=x0%d(Ck+2)*(Ck+1)*(Ck+2); ENDDO
    v=x*0.d0; v0=v; v1=v; 

    !Mathematical function phi
    KF=1; CALL ChebyshevTn_Real8(KF,n,x0%d(0),PL,DPL);  !KF=1, Tn
    v%d(0)=PL(n);
    KF=2; CALL ChebyshevTN_Real8(KF,N-1,x0%d(0),PL,DPL);  !KF=2, Tn
    v%d(1)=n*PL(n-1)*x%d(1);
    alpha=1-x0*x0;beta=-x0;
    gamma=n*n*1.d0;
 
    !Intermediate variables 
    x12=x1*x1; x13=x12*x1;
    d=1.d0/(alpha*x1);
    g=-gamma*x13*d; 
    a=(alpha*x2-beta*x12)*d; 

    !Generic AD of phi
    v0%d(0) = v%d(0);
    v0%d(1) = v%d(1); v1%d(0) =v%d(1);
    DO Ck=2,Ckmax
      v2=(0.d0,0.d0)
      DO j=0,Ck-2
        v2=v2+g%d(j)*v0%d(Ck-2-j)+a%d(j)*v1%d(Ck-2-j)
      ENDDO
      v0%d(Ck)=v2/((Ck-1)*Ck);   !unscaling
      v1%d(Ck-1)=v0%d(Ck)*Ck;
      v%d(Ck)=v0%d(Ck);
    ENDDO
    Ck=Ckmax 
  END SUBROUTINE ChebyshevTn_Real8d
END MODULE HOAD_ChebyshevTn

!Copyright (c) 2016, CNRS and University of Strasbourg, France
!All rights reserved.

!Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

!1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

!2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

!3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

!THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
