!Higher order automatic differentiation up to order Ck for the cosine function cos
!where Ck is a global variable defining the current truncation order of differentiation
!
!The HOAD_sin module allows to package the different versions of cos so that they can be imported into  program units to be automatically differentiated at higher order.
!
!cos is overloaded for Taylor series. Related HOAD types are defined in HOAD_types. 
!  * Real8d for double precision/real*8 Taylor coefficients 
!  * Complex16 for complex*16 Taylor coefficients 
!
!Author: I. Charpentier

!Note that the recurrence formula make use of both sine and cosine functions and
!are thus common to HOAD_cos.f90 and HOADsin.f90

!Formula: Griewank&Walther (2008) page 308

MODULE HOAD_cos
  USE HOAD_types
  IMPLICIT NONE
  INTERFACE COS
!Overloaded sin: the argument is a Taylor series (either Real8d or Complex16d)
    MODULE PROCEDURE cosReal8D
    MODULE PROCEDURE cosComplex16D
  END INTERFACE COS
CONTAINS 
!-----
!OVERLOADING c=cos(a): THE ARGUMENT IS A Real8d TAYLOR SERIES   
  FUNCTION cosReal8D(a) RESULT(c)
    TYPE(Real8D),INTENT(IN)::a
    TYPE(Real8D)::s  !Taylor coefficients for sine
    TYPE(Real8D)::c  !Taylor coefficients for cosine
    TYPE(Real8D)::t  !Scaled Taylor coefficients of a
    INTEGER j,l
!Recurrence formula for sine:   k*s_k=sum_{j=1}^k  (j*a_j)c_{k-j}, for all k
!Recurrence formula for cosine: k*c_k=sum_{j=1}^k -(j*a_j)s_{k-j}, for all k
    c%d(0) = COS(a%d(0)); s%d(0) = SIN(a%d(0))
    DO j = 1,Ck 
      t%d(j) = a%d(j)*j
      s%d(j) = 0.d0; c%d(j) = 0.d0
      DO l = 1,j
        s%d(j) = s%d(j) + t%d(l)*c%d(j-l)
        c%d(j) = c%d(j) - t%d(l)*s%d(j-l)
      ENDDO
      s%d(j) = s%d(j)/j !"unscaling" the sine   Taylor coefficient of order j
      c%d(j) = c%d(j)/j !"unscaling" the cosine Taylor coefficient of order j
    ENDDO
  END FUNCTION cosReal8D

!-----
!OVERLOADING c=cos(a): THE ARGUMENT IS A Complex16D TAYLOR SERIES   
  FUNCTION cosComplex16D(a) RESULT(c)
    TYPE(Complex16D),INTENT(IN)::a
    TYPE(Complex16D)::s  !Taylor coefficients for sine
    TYPE(Complex16D)::c  !Taylor coefficients for cosine
    TYPE(Complex16D)::t  !Scaled Taylor coefficients of a
    INTEGER j,l
!Recurrence formula for sine:   k*s_k=sum_{j=1}^k  (j*a_j)c_{k-j}, for all k
!Recurrence formula for cosine: k*c_k=sum_{j=1}^k -(j*a_j)s_{k-j}, for all k
    c%d(0) = COS(a%d(0)); s%d(0) = SIN(a%d(0))
    DO j = 1,Ck 
      t%d(j) = a%d(j)*j
      s%d(j) = 0.d0; c%d(j) = 0.d0
      DO l = 1,j
        s%d(j) = s%d(j) + t%d(l)*c%d(j-l)
        c%d(j) = c%d(j) - t%d(l)*s%d(j-l)
      ENDDO
      s%d(j) = s%d(j)/j !"unscaling" the sine   Taylor coefficient of order j
      c%d(j) = c%d(j)/j !"unscaling" the cosine Taylor coefficient of order j
    ENDDO
  END FUNCTION cosComplex16D
END MODULE HOAD_cos

!Copyright (c) 2016, CNRS and University of Strasbourg, France
!All rights reserved.

!Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

!1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

!2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

!3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

!THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
