!Higher order automatic differentiation up to order Ck for the exponential function exp
!where Ck is a global variable defining the current truncation order of differentiation
!
!The HOAD_exp module allows to package the different versions of exp so that they can be imported into  program units to be automatically differentiated at higher order.
!
!exp is overloaded for Taylor series. Related HOAD types are defined in HOAD_types. 
!  * Real8d for double precision/real*8 Taylor coefficients 
!  * Complex16 for complex*16 Taylor coefficients 
!
!Author: I. Charpentier

!Formula: Griewank&Walther (2008) page 308

MODULE HOAD_exp
  USE HOAD_TYPEs
  IMPLICIT NONE
  INTERFACE EXP
!Overloaded exp: the argument is a Taylor series (either Real8d or Complex16d)
    MODULE PROCEDURE exp_real8D
    MODULE PROCEDURE exp_complex16D
  END INTERFACE EXP
CONTAINS 
!-----
!OVERLOADING r=exp(a): THE ARGUMENT IS A Real8d TAYLOR SERIES 
  FUNCTION exp_real8D(a) RESULT(r)
    TYPE(real8d),INTENT(IN)::a
    TYPE(real8d)::r
    TYPE(real8d)::ta,tr !Scaled Taylor coefficients of a and r
    INTEGER j,k
!Recurrence formula: k*r_k=sum_{j=1}^Ck (j*a_j)*r_{k-j}, for all k
    r%d(0)=EXP(a%d(0))
    DO k = 1,Ck
      ta%d(k)=a%d(k)*k
      tr%d(k)=0.d0
      DO j = 1,k
        tr%d(k)= tr%d(k) + ta%d(j)*r%d(k-j)
      ENDDO
      r%d(k)= tr%d(k)/k
    ENDDO
  END FUNCTION exp_real8D
!-----
!OVERLOADING r=exp(a): THE ARGUMENT IS A Complex16D TAYLOR SERIES 
  FUNCTION exp_complex16D(a) RESULT(r)
    TYPE(complex16D),INTENT(IN)::a
    TYPE(complex16D)::r
    TYPE(complex16D)::ta,tr !Scaled Taylor coefficients of a and r
    INTEGER :: j,k
!Recurrence formula: k*r_k=sum_{j=1}^Ck (j*a_j)*r_{k-j}, for all k
    r%d(0)=EXP(a%d(0))
    DO k = 1,Ck
      ta%d(k)=a%d(k)*k
      tr%d(k)=0.d0
      DO j = 1,k
        tr%d(k)= tr%d(k) + ta%d(j)*r%d(k-j)
      ENDDO
      r%d(k)= tr%d(k)/k
    ENDDO
  END FUNCTION exp_complex16D
END MODULE HOAD_exp

!Copyright (c) 2016, CNRS and University of Strasbourg, France
!All rights reserved.

!Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

!1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

!2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

!3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

!THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
