!Higher order automatic differentiation up to order Ck for the power operator **
!where Ck is a global variable defining the current truncation order of differentiation
!
!The HOAD_power module allows to package the different versions of ** so that they can be imported into  program units to be automatically differentiated at higher order.
!
!** is overloaded for Taylor series. Related HOAD types are defined in HOAD_types. 
!  * Real8d for double precision/real*8 Taylor coefficients 
!  * Complex16 for complex*16 Taylor coefficients 
!
!Author: I. Charpentier

!The general formula is provided in Griewank&Walther (2008) page 308

MODULE HOAD_power
  USE HOAD_types
  IMPLICIT NONE
  INTERFACE OPERATOR(**)
!Overloaded **: the first operand is a Taylor series (either Real8d or Complex16d), the power is an integer or a real number
    MODULE PROCEDURE power_Real8d_Integer
    MODULE PROCEDURE power_Real8d_Real8
    MODULE PROCEDURE power_Complex16d_Integer
    MODULE PROCEDURE power_Complex16d_Real8
  END INTERFACE OPERATOR(**)
CONTAINS 
!-----
!OVERLOADING r=a**b: THE FIRST OPERAND IS A Real8d TAYLOR SERIES, THE POWER IS AN INTEGER
  FUNCTION power_Real8d_Integer(a,b) result(r)
    TYPE(Real8d),INTENT(IN)::a
    INTEGER, INTENT(IN) :: b
    TYPE(Real8d)::r,ta,tr
    INTEGER::k,j,l
    IF (b.EQ.2) THEN
    !Recurrence formula: r_k=sum_{j=0}^Ck a_j*a_{k-j}, for all k
      DO j = 0,Ck 
        r%d(j) = 0.d0
        DO l = 0,j
          r%d(j) = r%d(j) + a%d(l)*a%d(j-l)
        ENDDO
      ENDDO
    ELSE
      WRITE(*,*) 'POWER INTEGER: TO BE IMPLEMENTED'
    ENDIF
  END FUNCTION power_Real8d_Integer
!-----
!OVERLOADING r=a**b: THE FIRST OPERAND IS A Real8d TAYLOR SERIES, THE POWER IS AN REAL*8
  FUNCTION power_Real8d_Real8(a,b) result(r)
    TYPE(Real8d),INTENT(IN)::a
    REAL*8, INTENT(IN) :: b
    TYPE(Real8d)::r,ta,tr
    REAL*8:: recip
    INTEGER::k,j,l
    IF (b.EQ.2) THEN
    !Recurrence formula: r_k=sum_{j=0}^Ck a_j*a_{k-j}, for all k
      DO j = 0,Ck 
        r%d(j) = 0.d0
        DO l = 0,j
          r%d(j) = r%d(j) + a%d(l)*a%d(j-l)
        ENDDO
      ENDDO
    ELSE
      r%d(0)=a%d(0)**b
      recip=1.d0/a%d(0)
      DO k = 1,Ck
        ta%d(k)=a%d(k)*k
        tr%d(k)=a%d(k)*0
        DO j = 1,k
          tr%d(k)= tr%d(k) + r%d(k-j)*ta%d(j) 
        ENDDO
        tr%d(k)=b*tr%d(k)
        DO j = 1,k-1
          tr%d(k)= tr%d(k) - a%d(k-j)*tr%d(j) 
        ENDDO
        tr%d(k)=recip*tr%d(k)
        r%d(k)= tr%d(k)/k
      ENDDO
    ENDIF
  END FUNCTION power_Real8d_Real8
!-----
!OVERLOADING r=a**b: THE FIRST OPERAND IS A Complex16D TAYLOR SERIES, THE POWER IS AN INTEGER
  FUNCTION power_Complex16d_Integer(a,b) result(r)
    TYPE(Complex16d),INTENT(IN)::a
    INTEGER, INTENT(IN)::b
    TYPE(Complex16d)::r,ta,tr
    INTEGER:: k,j,l
    IF (b.EQ.2) THEN
    !Recurrence formula: r_k=sum_{j=0}^Ck a_j*a_{k-j}, for all k
      DO j = 0,Ck 
        r%d(j) = 0.d0
        DO l = 0,j
          r%d(j) = r%d(j) + a%d(l)*a%d(j-l)
        ENDDO
      ENDDO
    ELSE
      WRITE(*,*) 'POWER INTEGER: TO BE IMPLEMENTED'
    ENDIF 
  END FUNCTION power_Complex16d_Integer
!-----
!OVERLOADING r=a**b: THE FIRST OPERAND IS A Complex16D TAYLOR SERIES, THE POWER IS AN INTEGER
  FUNCTION power_Complex16d_Real8(a,b) result(r)
    TYPE(Complex16d),INTENT(IN)::a
    REAL*8, INTENT(IN)::b
    TYPE(Complex16d)::r,ta,tr
    COMPLEX*16:: recip
    INTEGER:: k,j,l
    IF (b.EQ.2) THEN
    !Recurrence formula: r_k=sum_{j=0}^Ck a_j*a_{k-j}, for all k
      DO j = 0,Ck 
        r%d(j) = 0.d0
        DO l = 0,j
          r%d(j) = r%d(j) + a%d(l)*a%d(j-l)
        ENDDO
      ENDDO
    ELSE
      r%d(0)=a%d(0)**b
      recip=1.d0/a%d(0)
      DO k = 1,Ck
        ta%d(k)=a%d(k)*k
        tr%d(k)=a%d(k)*0
        DO j = 1,k
          tr%d(k)= tr%d(k) + r%d(k-j)*ta%d(j) 
        ENDDO
        tr%d(k)=b*tr%d(k)
        DO j = 1,k-1
          tr%d(k)= tr%d(k) - a%d(k-j)*tr%d(j) 
        ENDDO
        tr%d(k)=recip*tr%d(k)
        r%d(k)= tr%d(k)/k
      ENDDO
    ENDIF 
  END FUNCTION power_Complex16d_Real8
END MODULE HOAD_power

!Copyright (c) 2016, CNRS and University of Strasbourg, France
!All rights reserved.

!Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

!1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

!2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

!3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

!THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
