!Higher order automatic differentiation up to order Ck for the multiplication operator *
!where Ck is a global variable defining the current truncation order of differentiation
!
!The HOAD_prod module allows to package the different versions of * so that they can be imported into  program units to be automatically differentiated at higher order.
!
!* is overloaded for Taylor series. Related HOAD types are defined in HOAD_types. 
!  * Real8d for double precision/real*8 Taylor coefficients 
!  * Complex16 for complex*16 Taylor coefficients 
!
!Author: I. Charpentier

MODULE HOAD_prod
  USE HOAD_types
  IMPLICIT NONE
!Combination of types for the overloaded *
!Convention: 
!  prod for *, 
!  first type (INTEGER, REAL*8, COMPLEX*16, Real8d, Complex16d)
!  second type (INTEGER, REAL*8, COMPLEX*16, Real8d, Complex16d)
!  HOAD combinations include at least Real8d or Complex16d
  INTERFACE OPERATOR(*)
!Overloaded *: the two operands are Taylor series
!4 combinations
    MODULE procedure prod_Real8d_Real8d
    MODULE procedure prod_Real8d_Complex16d
    MODULE procedure prod_Complex16d_Real8d
    MODULE procedure prod_Complex16d_Complex16d
!Overloaded *: the first operand is a Taylor series, but not the second one
!6 combinations
    MODULE procedure prod_Real8d_INTEGER
    MODULE procedure prod_Real8d_REAL8
    MODULE procedure prod_Real8d_COMPLEX16
    MODULE procedure prod_Complex16d_INTEGER
    MODULE procedure prod_Complex16d_REAL8
    MODULE procedure prod_Complex16d_COMPLEX16
!Overloaded *: the second operand is a Taylor series, but not the first one
!6 combinations
    MODULE procedure prod_INTEGER_Real8d
    MODULE procedure prod_INTEGER_Complex16d
    MODULE procedure prod_REAL8_Real8d
    MODULE procedure prod_REAL8_Complex16d
    MODULE procedure prod_COMPLEX16_Real8d
    MODULE procedure prod_COMPLEX16_Complex16d
  END INTERFACE OPERATOR(*)
CONTAINS 

!-----
!OVERLOADING r=a*b: THE TWO OPERANDS ARE TAYLOR SERIES (Real8d or Complex16d)
  FUNCTION prod_Real8d_Real8d(a,b) RESULT(r)
    TYPE(Real8d),INTENT(IN)::a
    TYPE(Real8d),INTENT(IN)::b
    TYPE(Real8d)::r
    INTEGER::j,k
!Recurrence formula: r_k=sum_{j=0}^Ck a_j*b_{k-j}, for all k
    DO k = 0,Ck
      r%d(k) = 0.d0
      DO j = 0,k
       r%d(k) = r%d(k) + a%d(j)*b%d(k-j)
      ENDDO
    ENDDO
  END FUNCTION prod_Real8d_Real8d

!same recurrence formula for the other 3 combinations
  FUNCTION prod_Real8d_Complex16d(a,b) RESULT(r)
    TYPE(Real8d),INTENT(IN)::a
    TYPE(Complex16d),INTENT(IN)::b
    TYPE(Complex16d)::r
    INTEGER::j,k
    DO k = 0,Ck
      r%d(k) = 0.d0
      DO j = 0,k
       r%d(k) = r%d(k) + a%d(j)*b%d(k-j)
      ENDDO
    ENDDO
  END FUNCTION prod_Real8d_Complex16d
  FUNCTION prod_Complex16d_Real8d(a,b) RESULT(r)
    TYPE(Complex16d),INTENT(IN)::a
    TYPE(Real8d),INTENT(IN)::b
    TYPE(Complex16d)::r
    INTEGER::j,k
    DO k = 0,Ck
      r%d(k) = 0.d0
      DO j = 0,k
       r%d(k) = r%d(k) + a%d(j)*b%d(k-j)
      ENDDO
    ENDDO
  END FUNCTION prod_Complex16d_Real8d
  FUNCTION prod_Complex16d_Complex16d(a,b) RESULT(r)
    TYPE(Complex16d),INTENT(IN)::a
    TYPE(Complex16d),INTENT(IN)::b
    TYPE(Complex16d)::r
    INTEGER::j,k
    DO k = 0,Ck
      r%d(k) = 0.d0
      DO j = 0,k
       r%d(k) = r%d(k) + a%d(j)*b%d(k-j)
      ENDDO
    ENDDO
  END FUNCTION prod_Complex16d_Complex16d

!-----
!OVERLOADING r=a*b: THE 1st OPERAND IS A TAYLOR SERIES (Real8d or Complex16d), BUT NOT THE 2nd ONE
  FUNCTION prod_Real8d_INTEGER(a,b) RESULT(r)
    TYPE(Real8d),INTENT(IN)::a
    INTEGER,INTENT(IN)::b
    TYPE(Real8d)::r
!Recurrence formula: r_k=a_k*b, for all k
    r%d(0:Ck)=a%d(0:Ck)*b
  END FUNCTION prod_Real8d_INTEGER

!same recurrence formula for the other 5 combinations
  FUNCTION prod_Real8d_REAL8(a,b) RESULT(r)
    TYPE(Real8d),INTENT(IN)::a
    REAL*8,INTENT(IN)::b
    TYPE(Real8d)::r
    r%d(0:Ck)=a%d(0:Ck)*b
  END FUNCTION prod_Real8d_REAL8
  FUNCTION prod_Real8d_COMPLEX16(a,b) RESULT(r)
    TYPE(Real8d),INTENT(IN)::a
    COMPLEX*16,INTENT(IN)::b
    TYPE(Complex16d)::r
    r%d(0:Ck)=a%d(0:Ck)*b
  END FUNCTION prod_Real8d_COMPLEX16
  FUNCTION prod_Complex16d_INTEGER(a,b) RESULT(r)
    TYPE(Complex16d),INTENT(IN)::a
    INTEGER,INTENT(IN)::b
    TYPE(Complex16d)::r
    r%d(0:Ck)=a%d(0:Ck)*b
  END FUNCTION prod_Complex16d_INTEGER
  FUNCTION prod_Complex16d_REAL8(a,b) RESULT(r)
    TYPE(Complex16d),INTENT(IN)::a
    REAL*8,INTENT(IN)::b
    TYPE(Complex16d)::r
    r%d(0:Ck)=a%d(0:Ck)*b
  END FUNCTION prod_Complex16d_REAL8
  FUNCTION prod_Complex16d_COMPLEX16(a,b) RESULT(r)
    TYPE(Complex16d),INTENT(IN)::a
    COMPLEX*16,INTENT(IN)::b
    TYPE(Complex16d)::r
    r%d(0:Ck)=a%d(0:Ck)*b
  END FUNCTION prod_Complex16d_COMPLEX16

!-----
!OVERLOADING r=a*b: THE 2nd OPERAND IS A TAYLOR SERIES (Real8d or Complex16d), BUT NOT THE 1st ONE
  FUNCTION prod_INTEGER_Real8d(a,b) RESULT(r)
    INTEGER,INTENT(IN)::a
    TYPE(Real8d),INTENT(IN)::b
    TYPE(Real8d)::r
!Recurrence formula: r_k=a*b_k, for all k
    r%d(0:Ck)=a*b%d(0:Ck)
  END FUNCTION prod_INTEGER_Real8d

!same recurrence formula for the other 5 combinations
  FUNCTION prod_INTEGER_Complex16d(a,b) RESULT(r)
    INTEGER,INTENT(IN)::a
    TYPE(Complex16d),INTENT(IN)::b
    TYPE(Complex16d)::r
    r%d(0:Ck)=a*b%d(0:Ck)
  END FUNCTION prod_INTEGER_Complex16d
  FUNCTION prod_REAL8_Real8d(a,b) RESULT(r)
    REAL*8,INTENT(IN)::a
    TYPE(Real8d),INTENT(IN)::b
    TYPE(Real8d)::r
    r%d(0:Ck)=a*b%d(0:Ck)
  END FUNCTION prod_REAL8_Real8d
  FUNCTION prod_REAL8_Complex16d(a,b) RESULT(r)
    REAL*8,INTENT(IN)::a
    TYPE(Complex16d),INTENT(IN)::b
    TYPE(Complex16d)::r
    r%d(0:Ck)=a*b%d(0:Ck)
  END FUNCTION prod_REAL8_Complex16d
  FUNCTION prod_COMPLEX16_Real8d(a,b) RESULT(r)
    COMPLEX*16,INTENT(IN)::a
    TYPE(Real8d),INTENT(IN)::b
    TYPE(Complex16d)::r
    r%d(0:Ck)=a*b%d(0:Ck)
  END FUNCTION prod_COMPLEX16_Real8d
  FUNCTION prod_COMPLEX16_Complex16d(a,b) RESULT(r)
    COMPLEX*16,INTENT(IN)::a
    TYPE(Complex16d),INTENT(IN)::b
    TYPE(Complex16d)::r
    r%d(0:Ck)=a*b%d(0:Ck)
  END FUNCTION prod_COMPLEX16_Complex16d
END MODULE HOAD_prod

!Copyright (c) 2016, CNRS and University of Strasbourg, France
!All rights reserved.

!Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

!1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

!2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

!3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

!THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
