!Higher order automatic differentiation up to order Ck for the sqrt function 
!where Ck is a global variable defining the current truncation order of differentiation
!
!The HOAD_sqrt module allows to package the different versions of sqrt so that they can be imported into  program units to be automatically differentiated at higher order.
!
!sqrt is overloaded for Taylor series. Related HOAD types are defined in HOAD_types. 
!  * Real8d for double precision/real*8 Taylor coefficients 
!  * Complex16 for complex*16 Taylor coefficients 
!
!Author: I. Charpentier

!Formula: Griewank&Walther (2008) page 305

MODULE HOAD_SQRT
  USE HOAD_TYPEs
  IMPLICIT NONE
  INTERFACE SQRT
!Overloaded sqrt: the argument is a Taylor series (either Real8d or Complex16d)
    MODULE PROCEDURE sqrt_Real8D
    MODULE PROCEDURE sqrt_Complex16D
  END INTERFACE SQRT
CONTAINS 
!-----
!OVERLOADING r=sqrt(a): THE ARGUMENT IS A Real8d TAYLOR SERIES 
  FUNCTION sqrt_Real8D(a) RESULT(r)
    TYPE(real8d),INTENT(IN)::a
    TYPE(real8d)::r
    DOUBLE PRECISION::recip
    INTEGER j,k
!Recurrence formula: r_0=sqrt(a_0)
!                    r_k=(1/(2r_0))*(a_k-sum_{j=1}^{Ck-1} r_j*r_{k-j}), for all k>0
    r%d(0)=SQRT(a%d(0))
    DO k = 1,Ck
      recip = 0.5d0/r%d(0)
      r%d(k) = a%d(k)
      DO j = 1,k-1
        r%d(k) = r%d(k) - r%d(j)*r%d(k-j)
      ENDDO
      r%d(k) = r%d(k)*recip
    ENDDO
  END FUNCTION sqrt_Real8D
!-----
!OVERLOADING r=sqrt(a): THE ARGUMENT IS A Complex16D TAYLOR SERIES 
  FUNCTION sqrt_Complex16D(a) RESULT(r)
    TYPE(complex16D),INTENT(IN)::a
    TYPE(complex16D)::r
    COMPLEX*16::recip
    INTEGER j,k
!Recurrence formula: r_0=sqrt(a_0)
!                    r_k=(1/(2r_0))*(a_k-sum_{j=1}^{Ck-1} r_j*r_{k-j}), for all k>0
    r%d(0)=SQRT(a%d(0))
    DO k = 1,Ck
      recip = 0.5d0/r%d(0)
      r%d(k) = a%d(k)
      DO j = 1,k-1
        r%d(k) = r%d(k) - r%d(j)*r%d(k-j)
      ENDDO
      r%d(k) = r%d(k)*recip
    ENDDO
  END FUNCTION sqrt_complex16D

END MODULE HOAD_SQRT

!Copyright (c) 2016, CNRS and University of Strasbourg, France
!All rights reserved.

!Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

!1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

!2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

!3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

!THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
